//+------------------------------------------------------------------+
//|                                                   moexsymbol.mq5 |
//|                                    Copyright (c) 2025, Marketeer |
//|                          https://www.mql5.com/ru/users/marketeer |
//|               Example for https://www.mql5.com/ru/articles/16964 |
//+------------------------------------------------------------------+
#property script_show_inputs
#property description "Create custom symbol for a ticker from MOEX ISS, update rates and ticks online."

#include "moex2mql5.mqh"
#include "moexref.mqh"

input group "MOEX Navigator"
input MOEX_ENGINES UseEngine = stock;
input MOEX_MARKETS UseMarket = stock·shares;
input MOEX_BOARDS UseBoard = shares·TQBR;

input group "MOEX Parameters"
input string UseTicker = "AFLT";
input datetime From;
input datetime To;
input MOEX_INTERVAL Timeframe = M1;

input group "Auxiliary"
input bool Logging = false;
input bool Dumping = true;
input bool Test = false;
input uint Timeout = 5; // Timeout (seconds)
input uint Delay = 100; // Delay (between paging requests, ms)
input string Parameters = "iss.meta=off"; // Global Parameters (such as iss.meta=on|off, etc)
input string CustomSuffix = ".iss"; // Custom Symbol Suffix (i.e. ".iss")

void OnStart()
{
   iss.link.applySettings(MoexLink::Settings(Logging, Dumping, Test, Timeout, Delay, Parameters));

   // just for your information print out history range (market and engine are autodetected)
   JsValue *range = iss.engines.candleborders(UseTicker, UseBoard)._get();
   Moex::HistoryRange history[];
   PASS(Moex::GetCandleborders(range, history));
   if(Logging) ArrayPrint(history);

   bool isCustom = false;
   if(!SymbolExist(UseTicker + CustomSuffix, isCustom))
   {
      Print("Creating new custom symbol ", UseTicker + CustomSuffix);
      string market, engine;
      
      // Find market and engine names to suggest SYMBOL_TRADE_CALC_MODE and symbol path
      Moex::ResolveMoexNames(UseBoard, UseMarket, UseEngine, market, engine, !UseMarket || !UseEngine);
      
      // Single ticker extensive description from iss/securities/XYZ
      JsValue *about = iss.securities.security(UseTicker)._get();
     
      // Specs for ticker /iss/engines/[engine]/markets/[market]/boards/[board]/securities/XYZ
      JsValue *spec = iss.engines.security(UseTicker, UseBoard, UseMarket, UseEngine)._sort_column("SECID")._get();
      
      if(spec.t == JS_OBJECT) // not an error or null
      {
         if(about.t == JS_OBJECT) // augment specs with description
         {
            spec.merge(about, true);
         }
         
         if(Dumping) DumpJsonToFile(StringFormat("MOEX/symbol-%s.json", UseTicker), spec);
         
         // Create custom symbol at engine/market/ticker path
         if(Moex::CreateCustomSymbol(spec, UseTicker + CustomSuffix, market, engine))
         {
            Print("Successfully created");
            isCustom = true;
         }
         else
         {
            Print("Failed with error: ", _LastError);
         }
      }
   }
   
   if(isCustom)
   {
      datetime last = (datetime)SeriesInfoInteger(UseTicker + CustomSuffix, Moex::Interval2Timeframe(Timeframe), SERIES_LASTBAR_DATE);
      PrintFormat("Updating custom symbol %s from %s", UseTicker + CustomSuffix, (string)last);
      int timing = (int)(Timeout * 1000);
      
      if(!last)
      {
         JsValue *data = iss.engines.candles(UseTicker, UseBoard, UseMarket, UseEngine)._interval(Timeframe)._from(fmin(From, To))._till(fmax(From, To))._start(MOEX_ALL)._get();
         MqlRates rates[];
         PASS(Moex::GetRates(data, rates));
         if(Logging) ArrayPrint(rates);
         PRTF(CustomRatesReplace(UseTicker + CustomSuffix, NULL, LONG_MAX, rates));
      }
      else
      {
         const static int DAYLONG = 60 * 60 * 24;
         long limit = 0;
         datetime prevday = 0, prevlast = 0, lasttick = 0;
         long next = 0;
         int no = 0; // serial number of last known tick
         int tickCount = 0, rateCount = 0;
         
         while(!IsStopped())
         {
            ResetLastError();
            JsValue *data = iss.engines.candles(UseTicker, UseBoard, UseMarket, UseEngine)._interval(Timeframe)._from(last)._till(TimeCurrent() + DAYLONG)._start(!limit ? MOEX_ALL : limit)._get();
            MqlRates rates[];
            PASS(Moex::GetRates(data, rates));
            const int n = ArraySize(rates);
            if(n)
            {
               PASS(CustomRatesUpdate(UseTicker + CustomSuffix, rates) + 1);
               last = rates[n - 1].time;
               long delta = prevlast ? (last / 60 * 60 - prevlast / 60 * 60) / 60 : 0; // next M1 bar(s) 1+
               if(prevlast != last) rateCount += n;
               prevlast = last;
               last = last / DAYLONG * DAYLONG;
               if(last != prevday && prevday) limit = 0; // reset offset of M1 bars inside every new day if updating online
               else limit = fmax(limit + delta, n - 1);
               prevday = last; // remember last known day
            }

            if(no != -1)
            {
               Moex::Trades2TicksDefault t2t;
               JsValue *trades;
               MqlTick ticks[];
               if(!next)
               {
                  trades = iss.engines.trades(UseTicker, UseBoard, UseMarket, UseEngine)._reversed(YES)._limit(1)._get("iss.only=trades");
                  PASS(t2t.GetTicks(trades, ticks));
               }
               else
               {
                  trades = iss.engines.trades(UseTicker, UseBoard, UseMarket, UseEngine)._recno(next)._tradeno(next)._next_trade(YES)._get("iss.only=trades");
                  PASS(t2t.GetTicks(trades, ticks));
               }
               if(trades[0]["data"].size())
               {
                  no = t2t[Moex::TradesColumns::recno] > -1 ? t2t[Moex::TradesColumns::recno] : t2t[Moex::TradesColumns::tradeno];
                  PASS(no != -1);
                  next = trades[0]["data"][trades[0]["data"].size() - 1][no].get<long>();
               }

               const int m = ArraySize(ticks);
               
               if(m)
               {
                  JsValue *online = iss.engines.securities(UseBoard, UseMarket, UseEngine)._securities(UseTicker)._get("iss.only=marketdata");
                  Moex::ColumnsToProperties(online);
                  const double bid = online["marketdata"]["0"]["bid"].get<double>();
                  const double ask = online["marketdata"]["0"]["offer"].get<double>();
                  if(ask != 0.0 && bid != 0.0)
                  {
                     for(int i = 0; i < m; i++)
                     {
                        if(ask > ticks[i].last) ticks[i].ask = ask;
                        if(bid < ticks[i].last) ticks[i].bid = bid;
                     }
                  }

                  PASS(CustomTicksAdd(UseTicker + CustomSuffix, ticks) + 1);
                  tickCount += m;
                  lasttick = ticks[m - 1].time;
               }
            }
            Comment(StringFormat("Bars:%d; Ticks:%d; Last time: %s", rateCount, tickCount, (string)(lasttick ? lasttick : prevlast)));
            if(_LastError == ERR_USER_ERROR_FIRST) // assert triggered
            {
               timing += timing;
               PrintFormat("Request tempo is slowing down: %ds", timing / 1000);
            }
            else
            {
               timing = (int)(Timeout * 1000);
            }
            Sleep(timing);
            Moex::Base::_fetch(NULL); // clean up cache once in a while (if objects are processed and not needed anymore)
         }
      }
      Comment("");
   }
}

//+------------------------------------------------------------------+
